"""
pkg1
pkg1.test
pkg1.test.test_things
pkg1.test.test_things.test_func
pkg1.test.test_things.test_gen
pkg1.test.test_things.test_gen:3
pkg1.test.test_things.SomeTests
pkg1.test.test_things.SomeTests.test_ok

# generator method
# generator method index

# param func
# param func index
# param method
# param method index

"""
from nose2.tests._common import FunctionalTestCase, support_file


class TestLoadTestsFromPackage(FunctionalTestCase):
    def test_start_directory_inside_package(self):
        proc = self.runIn(
            "scenario/tests_in_package/pkg1/test",
            "-v",
            "-t",
            support_file("scenario/tests_in_package"),
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 25 tests")
        self.assertEqual(proc.poll(), 1)

    def test_project_directory_inside_package(self):
        proc = self.runIn("scenario/tests_in_package/pkg1/test", "-v")
        self.assertTestRunOutputMatches(proc, stderr="Ran 25 tests")
        self.assertEqual(proc.poll(), 1)

    def test_module_name(self):
        proc = self.runIn("scenario/tests_in_package", "-v", "pkg1.test.test_things")
        self.assertTestRunOutputMatches(proc, stderr="Ran 25 tests")
        self.assertEqual(proc.poll(), 1)

    def test_package_name(self):
        proc = self.runIn("scenario/tests_in_package", "-v", "pkg1")
        self.assertTestRunOutputMatches(proc, stderr="Ran 25 tests")
        self.assertEqual(proc.poll(), 1)

    def test_module_name_with_start_dir(self):
        proc = self.runIn(
            ".",
            "-v",
            "-s",
            support_file("scenario/tests_in_package"),
            "pkg1.test.test_things",
        )

        self.assertTestRunOutputMatches(proc, stderr="Ran 25 tests")
        self.assertEqual(proc.poll(), 1)

    def test_package_name_with_start_dir(self):
        proc = self.runIn(
            ".", "-v", "-s", support_file("scenario/tests_in_package"), "pkg1"
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 25 tests")
        self.assertEqual(proc.poll(), 1)

    def test_function_name(self):
        proc = self.runIn(
            "scenario/tests_in_package", "-v", "pkg1.test.test_things.test_func"
        )
        self.assertTestRunOutputMatches(proc, stderr="test_func")
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertTestRunOutputMatches(proc, stderr="OK")
        self.assertEqual(proc.poll(), 0)

    def test_function_string_repr(self):
        proc = self.runIn(
            "scenario/tests_in_package", "-v", "pkg1.test.test_things.test_func"
        )
        self.assertTestRunOutputMatches(
            proc, stderr=r"pkg1.test.test_things.test_func \.\.\. ok"
        )
        self.assertEqual(proc.poll(), 0)

    def test_generator_function_name(self):
        proc = self.runIn(
            "scenario/tests_in_package", "-v", "pkg1.test.test_things.test_gen"
        )
        self.assertTestRunOutputMatches(proc, stderr="test_gen")
        self.assertTestRunOutputMatches(proc, stderr="Ran 5 tests")
        self.assertEqual(proc.poll(), 0)

    def test_generator_function_index(self):
        proc = self.runIn(
            "scenario/tests_in_package", "-v", "pkg1.test.test_things.test_gen:3"
        )
        self.assertTestRunOutputMatches(proc, stderr="test_gen")
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertEqual(proc.poll(), 0)

    def test_generator_function_index_1_based(self):
        proc = self.runIn(
            "scenario/tests_in_package", "-v", "pkg1.test.test_things.test_gen:1"
        )

        self.assertTestRunOutputMatches(proc, stderr="test_gen")
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertTestRunOutputMatches(proc, stderr="OK")
        self.assertEqual(proc.poll(), 0)

    def test_testcase_name(self):
        proc = self.runIn(
            "scenario/tests_in_package", "-v", "pkg1.test.test_things.SomeTests"
        )
        self.assertTestRunOutputMatches(proc, stderr="SomeTests")
        self.assertTestRunOutputMatches(proc, stderr="Ran 8 tests")
        self.assertEqual(proc.poll(), 1)

    def test_testcase_method(self):
        proc = self.runIn(
            "scenario/tests_in_package", "-v", "pkg1.test.test_things.SomeTests.test_ok"
        )
        self.assertTestRunOutputMatches(proc, stderr="SomeTests")
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertTestRunOutputMatches(proc, stderr="OK")
        self.assertEqual(proc.poll(), 0)

    def test_generator_method(self):
        proc = self.runIn(
            "scenario/tests_in_package",
            "-v",
            "pkg1.test.test_things.SomeTests.test_gen_method",
        )
        self.assertTestRunOutputMatches(proc, stderr="test_gen_method")
        self.assertTestRunOutputMatches(proc, stderr="Ran 2 tests")
        self.assertEqual(proc.poll(), 1)

    def test_generator_method_index(self):
        proc = self.runIn(
            "scenario/tests_in_package",
            "-v",
            "pkg1.test.test_things.SomeTests.test_gen_method:1",
        )
        self.assertTestRunOutputMatches(proc, stderr="test_gen_method")
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertTestRunOutputMatches(proc, stderr="OK")
        self.assertEqual(proc.poll(), 0)

    def test_parameterized_method(self):
        proc = self.runIn(
            "scenario/tests_in_package",
            "-v",
            "pkg1.test.test_things.SomeTests.test_params_method",
        )
        self.assertTestRunOutputMatches(proc, stderr="test_params_method")
        self.assertTestRunOutputMatches(proc, stderr="Ran 2 tests")
        self.assertEqual(proc.poll(), 1)

    def test_parameterized_method_index(self):
        proc = self.runIn(
            "scenario/tests_in_package",
            "-v",
            "pkg1.test.test_things.SomeTests.test_params_method:1",
        )
        self.assertTestRunOutputMatches(proc, stderr="test_params_method")
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertTestRunOutputMatches(proc, stderr="OK")
        self.assertEqual(proc.poll(), 0)

    def test_parameterized_func(self):
        proc = self.runIn(
            "scenario/tests_in_package", "-v", "pkg1.test.test_things.test_params_func"
        )
        self.assertTestRunOutputMatches(proc, stderr="test_params_func")
        self.assertTestRunOutputMatches(proc, stderr="Ran 2 tests")
        self.assertEqual(proc.poll(), 1)

    def test_parameterized_func_index(self):
        proc = self.runIn(
            "scenario/tests_in_package",
            "-v",
            "pkg1.test.test_things.test_params_func:1",
        )
        self.assertTestRunOutputMatches(proc, stderr="test_params_func")
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertTestRunOutputMatches(proc, stderr="OK")
        self.assertEqual(proc.poll(), 0)


class TestLoadTestsOutsideOfPackage(FunctionalTestCase):
    def test_module_name(self):
        proc = self.runIn("scenario/package_in_lib", "-v", "tests")
        self.assertTestRunOutputMatches(proc, stderr="Ran 3 tests")
        self.assertEqual(proc.poll(), 1)

    def test_function_name(self):
        proc = self.runIn("scenario/package_in_lib", "-v", "tests.test")
        self.assertTestRunOutputMatches(proc, stderr="test")
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertTestRunOutputMatches(proc, stderr="OK")
        self.assertEqual(proc.poll(), 0)

    def test_module_name_with_start_dir(self):
        proc = self.runIn(
            ".", "-v", "-s", support_file("scenario/package_in_lib"), "tests"
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 3 tests")
        self.assertEqual(proc.poll(), 1)


class TestLoadingErrors(FunctionalTestCase):
    def test_import_error_module(self):
        proc = self.runIn("scenario/module_import_err", "-v", "test_import_err")
        self.assertTestRunOutputMatches(proc, stderr="ImportError: booms")
        self.assertTestRunOutputMatches(
            proc, stderr=r"Traceback \(most recent call last\):"
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertEqual(proc.poll(), 1)

    def test_import_error_func(self):
        proc = self.runIn("scenario/module_import_err", "-v", "test_import_err.test")
        self.assertTestRunOutputMatches(proc, stderr="ImportError: booms")
        self.assertTestRunOutputMatches(
            proc, stderr=r"Traceback \(most recent call last\):"
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertEqual(proc.poll(), 1)

    def test_import_error_testcase(self):
        proc = self.runIn("scenario/module_import_err", "-v", "test_import_err.Test")
        self.assertTestRunOutputMatches(proc, stderr="ImportError: booms")
        self.assertTestRunOutputMatches(
            proc, stderr=r"Traceback \(most recent call last\):"
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertEqual(proc.poll(), 1)

    def test_import_error_testcase_method(self):
        proc = self.runIn(
            "scenario/module_import_err", "-v", "test_import_err.Test.test"
        )
        self.assertTestRunOutputMatches(proc, stderr="ImportError: booms")
        self.assertTestRunOutputMatches(
            proc, stderr=r"Traceback \(most recent call last\):"
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertEqual(proc.poll(), 1)

    def test_import_error_package(self):
        proc = self.runIn("scenario/module_import_err", "-v", "pkg")
        # In this case, there should not be an AttributeError since we only
        # import pkg, which should work.
        self.assertTestRunOutputMatches(
            proc,
            stderr="ImportError: Failed to import test module: pkg.test_import_err",
        )
        self.assertTestRunOutputMatches(
            proc, stderr=r"Traceback \(most recent call last\):"
        )
        self.assertTestRunOutputMatches(proc, stderr="ValueError: booms")
        # 3 tests should run, and only one should fail. We're testing that a
        # loading error does not prevent from running other tests.
        self.assertTestRunOutputMatches(proc, stderr="Ran 3 tests")
        self.assertTestRunOutputMatches(proc, stderr=r"FAILED \(errors=1\)")
        self.assertEqual(proc.poll(), 1)

    def test_import_error_module_in_package(self):
        proc = self.runIn("scenario/module_import_err", "-v", "pkg.test_import_err")
        # In this case, there should not be an ImportError. The import of
        # pkg.test_import_err fails due with a `ValueError`, and this is the
        # one we are expecting.
        self.assertTestRunOutputMatches(proc, stderr="AttributeError: ")
        self.assertTestRunOutputMatches(proc, stderr="ValueError: booms")
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertEqual(proc.poll(), 1)

    def test_import_error_unknown_module_in_package(self):
        proc = self.runIn("scenario/module_import_err", "-v", "pkg.test_does_not_exit")
        # In this case, there should not be an ImportError. The import of
        # pkg.test_import_err fails due with a `ValueError`, and this is the
        # one we are expecting.
        self.assertTestRunOutputMatches(proc, stderr="AttributeError: ")

        # py3.6 introduced ModuleNotFoundError, a subclass of ImportError
        importerr = "ModuleNotFoundError: No module named"
        self.assertTestRunOutputMatches(proc, stderr=importerr)

        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertEqual(proc.poll(), 1)


class TestTestClassLoading(FunctionalTestCase):
    def test_load_testclass_by_name(self):
        proc = self.runIn("scenario/test_classes", "-v", "test_classes.Test")
        self.assertTestRunOutputMatches(proc, stderr="Ran 8 tests")
        self.assertEqual(proc.poll(), 0)

    def test_load_testclass_method_by_name(self):
        proc = self.runIn("scenario/test_classes", "-v", "test_classes.Test.test")
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertEqual(proc.poll(), 0)

    def test_load_testclass_generator_method_by_name(self):
        proc = self.runIn("scenario/test_classes", "-v", "test_classes.Test.test_gen")
        self.assertTestRunOutputMatches(proc, stderr="Ran 5 tests")
        self.assertEqual(proc.poll(), 0)

    def test_load_testclass_params_method_by_name(self):
        proc = self.runIn(
            "scenario/test_classes", "-v", "test_classes.Test.test_params"
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 2 tests")
        self.assertEqual(proc.poll(), 0)

    def test_class_level_fixtures_supported(self):
        proc = self.runIn("scenario/test_classes", "-v", "test_fixtures")
        self.assertTestRunOutputMatches(proc, stderr="Ran 5 tests")
        self.assertEqual(proc.poll(), 0)

    def test_error_in_test_class(self):
        proc = self.runIn("scenario/test_class_fail", "-v", "test_class_fail")
        self.assertTestRunOutputMatches(proc, stderr="nose2.loader.LoadTestsFailure")
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertTestRunOutputMatches(proc, stderr="FAILED")
        self.assertEqual(proc.poll(), 1)

    def test_expected_failures(self):
        proc = self.runIn("scenario/expected_failures", "-v", "expected_failures")
        self.assertTestRunOutputMatches(
            proc,
            stderr=(
                r"FAILED \(failures=1, expected failures=1, unexpected successes=1\)"
            ),
        )



import unittest
from unittest import mock

from nose2 import events, loader, session
from nose2.plugins.loader import functions
from nose2.tests._common import TestCase


class TestFunctionLoader(TestCase):
    def setUp(self):
        self.session = session.Session()
        self.loader = loader.PluggableTestLoader(self.session)
        self.plugin = functions.Functions(session=self.session)

    def test_can_load_test_functions_from_module(self):
        class Mod:
            pass

        def test():
            pass

        m = Mod()
        m.test = test
        event = events.LoadFromModuleEvent(self.loader, m)
        self.session.hooks.loadTestsFromModule(event)
        self.assertEqual(len(event.extraTests), 1)
        assert isinstance(event.extraTests[0], unittest.FunctionTestCase)

    def test_ignores_generator_functions(self):
        class Mod:
            pass

        def test():
            yield

        m = Mod()
        m.test = test
        event = events.LoadFromModuleEvent(self.loader, m)
        self.session.hooks.loadTestsFromModule(event)
        self.assertEqual(len(event.extraTests), 0)

    def test_ignores_functions_that_take_args(self):
        class Mod:
            pass

        def test(a):
            pass

        m = Mod()
        m.test = test
        event = events.LoadFromModuleEvent(self.loader, m)
        self.session.hooks.loadTestsFromModule(event)
        self.assertEqual(len(event.extraTests), 0)

    def test_can_load_test_functions_from_name(self):
        event = events.LoadFromNameEvent(self.loader, __name__ + ".func", None)
        suite = self.session.hooks.loadTestsFromName(event)
        self.assertNotEqual(suite, None)

    def test_ignores_test_methods_from_name(self):
        # Should ignore test methods even when specified directly
        event = events.LoadFromNameEvent(
            self.loader, __name__ + ".Case.test_method", None
        )
        suite = self.session.hooks.loadTestsFromName(event)
        self.assertEqual(suite, None)

    def test_ignores_decorated_test_methods_from_name(self):
        # Should ignore test methods even when they are of FunctionType
        event = events.LoadFromNameEvent(
            self.loader, __name__ + ".Case.test_patched", None
        )
        suite = self.session.hooks.loadTestsFromName(event)
        self.assertEqual(suite, None)


def func():
    pass


def dummy():
    pass


class Case(unittest.TestCase):
    __test__ = False  # do not run this

    def test_method(self):
        pass

    @mock.patch(__name__ + ".dummy")
    def test_patched(self, mock):
        pass



#!/usr/bin/env python

from nose2.tools import such

with such.A("system") as it:
    with it.having("something"):

        @it.should("do stuff")
        def test_should_do_stuff(case):
            pass

    it.createTests(globals())

with such.A("another system") as it:
    with it.having("something"):

        @it.should("do stuff")  # noqa: F811
        def test_should_do_stuff(case):  # noqa: F811
            pass

    it.createTests(globals())



from nose2 import events, loader, session, util
from nose2.plugins.loader import generators, testcases
from nose2.tests._common import TestCase


class TestGeneratorUnpack(TestCase):
    tags = ["unit"]

    def setUp(self):
        self.session = session.Session()
        self.loader = loader.PluggableTestLoader(self.session)
        self.expect = [
            (0, ("call", (0, 1))),
            (1, ("call", (1, 2))),
            (2, ("call", (2, 3))),
        ]
        self.plugin = generators.Generators(session=self.session)
        # need testcase loader to make the initial response to load from module
        self.tcl = testcases.TestCaseLoader(session=self.session)

    def test_unpack_handles_nose_style_generators(self):
        def gen():
            for i in range(0, 3):
                yield "call", i, i + 1

        out = list(self.plugin.unpack(gen()))
        self.assertEqual(out, self.expect)

    def test_unpack_handles_unittest2_style_generators(self):
        def gen():
            for i in range(0, 3):
                yield "call", (i, i + 1)

        out = list(self.plugin.unpack(gen()))
        self.assertEqual(out, self.expect)

    def test_ignores_ordinary_functions(self):
        class Mod:
            pass

        def test():
            pass

        m = Mod()
        m.test = test
        event = events.LoadFromModuleEvent(self.loader, m)
        self.session.hooks.loadTestsFromModule(event)
        self.assertEqual(len(event.extraTests), 0)

    def test_can_load_tests_from_generator_functions(self):
        class Mod:
            __name__ = "themod"

        def check(x):
            assert x == 1

        def test():
            yield check, 1
            yield check, 2

        m = Mod()
        m.test = test
        test.__module__ = m.__name__
        event = events.LoadFromModuleEvent(self.loader, m)
        self.session.hooks.loadTestsFromModule(event)
        self.assertEqual(len(event.extraTests), 2)
        # check that test names are sensible
        self.assertEqual(util.test_name(event.extraTests[0]), "themod.test:1")
        self.assertEqual(util.test_name(event.extraTests[1]), "themod.test:2")

    def test_can_load_tests_from_generator_methods(self):
        class Mod:
            pass

        def check(x):
            return x == 1

        class Test(TestCase):
            def test(self):
                yield check, 1
                yield check, 2

        m = Mod()
        m.Test = Test
        event = events.LoadFromModuleEvent(self.loader, m)
        self.session.hooks.loadTestsFromModule(event)
        self.assertEqual(len(event.extraTests), 1)
        self.assertEqual(len(event.extraTests[0]._tests), 2)



import os
from xml.etree import ElementTree as ET

from nose2.tests._common import FunctionalTestCase, TestCase, _method_name, support_file


class JunitXmlPluginFunctionalTest(FunctionalTestCase, TestCase):
    _RUN_IN_TEMP = True

    def run_with_junitxml_loaded(self, scenario, *args, **kwargs):
        work_dir = os.getcwd()
        test_dir = support_file(*scenario)
        junit_report = os.path.join(
            work_dir, kwargs.get("junit_report", "nose2-junit.xml")
        )
        config = os.path.join(test_dir, "unittest.cfg")
        config_args = ()
        if os.path.exists(junit_report):
            os.remove(junit_report)
        if os.path.exists(config):
            config_args = ("-c", config)
        proc = self.runIn(
            work_dir,
            "-s%s" % test_dir,
            "--plugin=nose2.plugins.junitxml",
            "-v",
            *(config_args + args),
        )
        return junit_report, proc

    def test_invocation_by_double_dash_option(self):
        junit_report, proc = self.run_with_junitxml_loaded(
            ("scenario", "junitxml", "happyday"), "--junit-xml"
        )

        self.assertTestRunOutputMatches(
            proc,
            stderr=r"test \(test_junitxml_happyday.Test"
            + _method_name()
            + r"\) ... ok",
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertEqual(proc.poll(), 0)

        self.assertTrue(
            os.path.isfile(junit_report),
            "junitxml report wasn't found in working directory. "
            "Searched for " + junit_report,
        )

    def test_invocation_by_single_dash_option(self):
        junit_report, proc = self.run_with_junitxml_loaded(
            ("scenario", "junitxml", "happyday"), "-X"
        )

        self.assertTestRunOutputMatches(
            proc,
            stderr=r"test \(test_junitxml_happyday.Test"
            + _method_name()
            + r"\) ... ok",
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertEqual(proc.poll(), 0)

        self.assertTrue(
            os.path.isfile(junit_report),
            "junitxml report wasn't found in working directory. "
            "Searched for " + junit_report,
        )

    def test_implicit_registration_by_path_option(self):
        junit_report, proc = self.run_with_junitxml_loaded(
            ("scenario", "junitxml", "happyday"),
            "--junit-xml-path=b.xml",
            junit_report="b.xml",
        )

        self.assertTestRunOutputMatches(
            proc,
            stderr=r"test \(test_junitxml_happyday.Test"
            + _method_name()
            + r"\) ... ok",
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertEqual(proc.poll(), 0)

        self.assertTrue(
            os.path.isfile(junit_report),
            "junitxml report wasn't found in working directory. "
            "Searched for " + junit_report,
        )

    def test_no_report_written_if_loaded_but_not_invoked(self):
        junit_report, proc = self.run_with_junitxml_loaded(
            ("scenario", "junitxml", "happyday")
        )

        self.assertTestRunOutputMatches(
            proc,
            stderr=r"test \(test_junitxml_happyday.Test"
            + _method_name()
            + r"\) ... ok",
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertEqual(proc.poll(), 0)

        self.assertFalse(
            os.path.isfile(junit_report),
            "junitxml report was found in working directory. "
            "Report file: " + junit_report,
        )

    def test_report_location_should_be_resilent_to_chdir_in_tests(self):
        junit_report, proc = self.run_with_junitxml_loaded(
            ("scenario", "junitxml", "chdir"), "--junit-xml"
        )

        self.assertTestRunOutputMatches(
            proc,
            stderr=r"test_chdir \(test_junitxml_chdir.Test"
            + _method_name("test_chdir")
            + r"\) \.* ok",
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertEqual(proc.poll(), 0)

        self.assertTrue(
            os.path.isfile(junit_report),
            "junitxml report wasn't found in working directory. "
            "Searched for " + junit_report,
        )

    def test_report_includes_properties(self):
        work_dir = os.getcwd()
        with open(os.path.join(work_dir, "properties.json"), "w") as fh:
            fh.write('{"PROPERTY_NAME":"PROPERTY_VALUE"}')
        junit_report, proc = self.run_with_junitxml_loaded(
            ("scenario", "junitxml", "with_properties"), "--junit-xml"
        )
        self.assertTestRunOutputMatches(
            proc,
            stderr=r"test \(test_junitxml_with_properties.Test"
            + _method_name()
            + r"\) \.* ok",
        )
        self.assertEqual(proc.poll(), 0)

        with open(junit_report) as fh:
            tree = ET.parse(fh).getroot()
        self.assertEqual(len(tree.findall("properties")), 1)
        prop = tree.find("properties").find("property")
        assert "name" in prop.attrib
        assert "value" in prop.attrib
        self.assertEqual(prop.get("name"), "PROPERTY_NAME")
        self.assertEqual(prop.get("value"), "PROPERTY_VALUE")

    def test_skip_reason_in_message(self):
        junit_report, proc = self.run_with_junitxml_loaded(
            ("scenario", "junitxml", "skip_reason"), "--junit-xml"
        )

        self.assertTestRunOutputMatches(
            proc,
            stderr=r"test \(test_junitxml_skip_reason.Test"
            + _method_name()
            + r"\) \.* skip",
        )

        exit_status = proc.poll()
        assert exit_status == 0

        with open(junit_report) as fh:
            tree = ET.parse(fh).getroot()

        num_test_cases = len(tree.findall("testcase"))
        assert num_test_cases == 1

        num_skipped = len(tree.find("testcase").findall("skipped"))
        assert num_skipped == 1
        skip_node = tree.find("testcase").find("skipped")
        assert "message" in skip_node.attrib
        skip_message = skip_node.get("message")
        assert skip_message == "test skipped: ohai"

    def test_xml_path_override_by_config(self):
        junit_report, proc = self.run_with_junitxml_loaded(
            ("scenario", "junitxml", "non_default_path"),
            "--junit-xml",
            junit_report="a.xml",
        )

        self.assertTestRunOutputMatches(
            proc,
            stderr=r"test \(test_junitxml_non_default_path.Test"
            + _method_name()
            + r"\) \.* ok",
        )

        exit_status = proc.poll()
        assert exit_status == 0

        self.assertTrue(os.path.isfile(junit_report))

    def test_xml_path_override_by_command(self):
        junit_report, proc = self.run_with_junitxml_loaded(
            ("scenario", "junitxml", "non_default_path"),
            "--junit-xml",
            "--junit-xml-path=b.xml",
            junit_report="b.xml",
        )

        self.assertTestRunOutputMatches(
            proc,
            stderr=r"test \(test_junitxml_non_default_path.Test"
            + _method_name()
            + r"\) \.* ok",
        )

        exit_status = proc.poll()
        assert exit_status == 0

        self.assertTrue(os.path.isfile(junit_report))


class JunitXmlPluginFunctionalFailureTest(FunctionalTestCase, TestCase):
    def test_failure_to_write_report(self):
        proc = self.runIn(
            "scenario/junitxml/fail_to_write",
            "--plugin=nose2.plugins.junitxml",
            "-v",
            "--junit-xml",
        )
        self.assertEqual(proc.poll(), 1)

        self.assertTestRunOutputMatches(
            proc,
            stderr=r"test \(test_junitxml_fail_to_write.Test"
            + _method_name()
            + r"\) \.* ok",
        )

        filename_for_regex = os.path.abspath("/does/not/exist.xml")
        filename_for_regex = filename_for_regex.replace("\\", r"\\\\")
        self.assertTestRunOutputMatches(
            proc,
            stderr=r"Internal Error: runTests aborted: \[Errno 2\] "
            "JUnitXML: Parent folder does not exist for file: "
            "'%s'" % filename_for_regex,
        )

    def test_failure_to_read_missing_properties(self):
        proc = self.runIn(
            "scenario/junitxml/missing_properties",
            "--plugin=nose2.plugins.junitxml",
            "-v",
            "--junit-xml",
        )
        self.assertEqual(proc.poll(), 1)

        self.assertTestRunOutputMatches(
            proc,
            stderr=r"test \(test_junitxml_missing_properties.Test"
            + _method_name()
            + r"\) \.* ok",
        )

        filename_for_regex = os.path.join("missing_properties", "properties.json")
        filename_for_regex = filename_for_regex.replace("\\", r"\\\\")
        self.assertTestRunOutputMatches(
            proc,
            stderr=r"Internal Error: runTests aborted: \[Errno 2\] "
            "JUnitXML: Properties file does not exist: "
            "'.*%s'" % filename_for_regex,
        )

    def test_failure_to_read_empty_properties(self):
        proc = self.runIn(
            "scenario/junitxml/empty_properties",
            "--plugin=nose2.plugins.junitxml",
            "-v",
            "--junit-xml",
        )
        self.assertEqual(proc.poll(), 1)

        self.assertTestRunOutputMatches(
            proc,
            stderr=r"test \(test_junitxml_empty_properties.Test"
            + _method_name()
            + r"\) \.* ok",
        )

        filename_for_regex = os.path.join("empty_properties", "properties.json")
        filename_for_regex = filename_for_regex.replace("\\", r"\\")
        self.assertTestRunOutputMatches(
            proc,
            stderr="Internal Error: runTests aborted: "
            "JUnitXML: could not decode file: "
            "'.*%s'" % filename_for_regex,
        )



from nose2 import events


class PrintFixture(events.Plugin):
    alwaysOn = True

    def startLayerSetup(self, event):
        print(f"StartLayerSetup: {event.layer}")

    def stopLayerSetup(self, event):
        print(f"StopLayerSetup: {event.layer}")

    def startLayerSetupTest(self, event):
        log = "StartLayerSetupTest: {0}:{1}"
        print(log.format(event.layer, event.test))

    def stopLayerSetupTest(self, event):
        log = "StopLayerSetupTest: {0}:{1}"
        print(log.format(event.layer, event.test))

    def startLayerTeardownTest(self, event):
        log = "StartLayerTeardownTest: {0}:{1}"
        print(log.format(event.layer, event.test))

    def stopLayerTeardownTest(self, event):
        log = "StopLayerTeardownTest: {0}:{1}"
        print(log.format(event.layer, event.test))

    def startLayerTeardown(self, event):
        print(f"StartLayerTeardown: {event.layer}")

    def stopLayerTeardown(self, event):
        print(f"StopLayerTeardown: {event.layer}")



from nose2.tools import such
from nose2.tools.params import params

with such.A("foo") as it:

    @it.should("do bar")
    @params(1, 2, 3)
    def test(case, bar):
        case.assertTrue(isinstance(bar, int))

    @it.should("do bar and extra")
    @params((1, 2), (3, 4), (5, 6))
    def testExtraArg(case, bar, foo):
        case.assertTrue(isinstance(bar, int))
        case.assertTrue(isinstance(foo, int))


it.createTests(globals())



"""
Output test reports in junit-xml format.

This plugin implements :func:`startTest`, :func:`testOutcome` and
:func:`stopTestRun` to compile and then output a test report in
junit-xml format. By default, the report is written to a file called
``nose2-junit.xml`` in the current working directory.

You can configure the output filename by setting ``path`` in a ``[junit-xml]``
section in a config file.  Unicode characters which are invalid in XML 1.0
are replaced with the ``U+FFFD`` replacement character. In the case that your
software throws an error with an invalid byte string.

By default, the ranges of discouraged characters are replaced as well. This can
be changed by setting the ``keep_restricted`` configuration variable to
``True``.

By default, the arguments of parametrized and generated tests are not printed.
For instance, the following code:

.. code-block:: python

    # a.py

    from nose2 import tools

    def test_gen():
        def check(a, b):
            assert a == b, '{}!={}'.format(a,b)

        yield check, 99, 99
        yield check, -1, -1

    @tools.params('foo', 'bar')
    def test_params(arg):
        assert arg in ['foo', 'bar', 'baz']

Produces this XML by default:

.. code-block:: xml

    <testcase classname="a" name="test_gen:1" time="0.000171"
     timestamp="2021-12-09T21:28:09.686611">
        <system-out />
    </testcase>
    <testcase classname="a" name="test_gen:2" time="0.000202"
     timestamp="2021-12-09T21:28:09.686813">
        <system-out />
    </testcase>
    <testcase classname="a" name="test_params:1" time="0.000159"
     timestamp="2021-12-09T21:28:09.686972">
        <system-out />
    </testcase>
    <testcase classname="a" name="test_params:2" time="0.000163"
     timestamp="2021-12-09T21:28:09.687135">
        <system-out />
    </testcase>

But if ``test_fullname`` is ``True``, then the following XML is
produced:

.. code-block:: xml

    <testcase classname="a" name="test_gen:1 (99, 99)" time="0.000213"
     timestamp="2021-12-09T21:28:09.686611">
        <system-out />
    </testcase>
    <testcase classname="a" name="test_gen:2 (-1, -1)" time="0.000194"
     timestamp="2021-12-09T21:28:09.687105">
        <system-out />
    </testcase>
    <testcase classname="a" name="test_params:1 ('foo')" time="0.000178"
     timestamp="2021-12-09T21:28:09.687283">
        <system-out />
    </testcase>
    <testcase classname="a" name="test_params:2 ('bar')" time="0.000187"
     timestamp="2021-12-09T21:28:09.687470">
        <system-out />
    </testcase>

"""
# Based on unittest2/plugins/junitxml.py,
# which is itself based on the junitxml plugin from py.test

import datetime
import json
import os.path
import re
import sys
import time
from xml.etree import ElementTree as ET

from nose2 import events, result, util

__unittest = True


class JUnitXmlReporter(events.Plugin):
    """Output junit-xml test report to file"""

    configSection = "junit-xml"
    commandLineSwitch = ("X", "junit-xml", "Generate junit-xml output report")

    def __init__(self):
        # Read argument from configuration file, or filled with default
        self.path = os.path.realpath(
            self.config.as_str("path", default="nose2-junit.xml")
        )
        self.keep_restricted = self.config.as_bool("keep_restricted", default=False)
        self.test_properties = self.config.as_str("test_properties", default=None)
        self.test_fullname = self.config.as_bool("test_fullname", default=False)

        if self.test_properties is not None:
            self.test_properties_path = os.path.realpath(self.test_properties)
        self.errors = 0
        self.failed = 0
        self.skipped = 0
        self.numtests = 0
        self.tree = ET.Element("testsuite")
        self._start = None

        # Allow user to override certain option from command line
        group = self.session.pluginargs
        group.add_argument(
            "--junit-xml-path",
            action="store",
            default="",
            metavar="FILE",
            dest="path",
            help="Output XML filename",
        )

    def handleArgs(self, event):
        """Read option from command line and override the value in config file
        when necessary"""
        if event.args.path:
            # register if `--junit-xml-path` is given
            # this avoids a situation in which a plugin-related option is given
            # but the plugin isn't activated, which can be confusing (see #521)
            self.register()
            self.path = os.path.realpath(event.args.path)

    def startTest(self, event):
        """Count test, record start time"""
        self.numtests += 1
        self._start = event.startTime

    def testOutcome(self, event):
        """Add test outcome to xml tree"""
        test = event.test
        testid_lines = test.id().split("\n")
        testid = testid_lines[0]
        parts = testid.split(".")
        classname = ".".join(parts[:-1])
        method = parts[-1]
        # for generated test cases
        if len(testid_lines) > 1 and self.test_fullname:
            test_args = ":".join(testid_lines[1:])
            method = f"{method} ({test_args})"

        # subtests do not report success
        if event.outcome == result.SUBTEST and event.exc_info is None:
            return

        testcase = ET.SubElement(self.tree, "testcase")
        testcase.set("timestamp", self._iso_timestamp())
        testcase.set("time", "%.6f" % self._time())
        if not classname:
            classname = test.__module__
        testcase.set("classname", classname)
        testcase.set("name", method)

        msg = ""
        if event.exc_info:
            msg = util.exc_info_to_string(event.exc_info, test)
        elif event.reason:
            msg = event.reason

        msg = string_cleanup(msg, self.keep_restricted)

        if event.outcome == result.ERROR:
            self.errors += 1
            error = ET.SubElement(testcase, "error")
            error.set("message", "test failure")
            error.text = msg
        elif event.outcome == result.FAIL and not event.expected:
            self.failed += 1
            failure = ET.SubElement(testcase, "failure")
            failure.set("message", "test failure")
            failure.text = msg
        elif event.outcome == result.PASS and not event.expected:
            self.skipped += 1
            skipped = ET.SubElement(testcase, "skipped")
            skipped.set("message", "test passes unexpectedly")
        elif event.outcome == result.SKIP:
            self.skipped += 1
            skipped = ET.SubElement(testcase, "skipped")
            if msg:
                skipmsg = "test skipped"
                if event.reason:
                    skipmsg = f"test skipped: {event.reason}"
                skipped.set("message", skipmsg)
                skipped.text = msg
        elif event.outcome == result.FAIL and event.expected:
            self.skipped += 1
            skipped = ET.SubElement(testcase, "skipped")
            skipped.set("message", "expected test failure")
            skipped.text = msg
        elif event.outcome == result.SUBTEST:
            if issubclass(event.exc_info[0], event.test.failureException):
                self.failed += 1
                failure = ET.SubElement(testcase, "failure")
                failure.set("message", "test failure")
                failure.text = msg
            else:
                self.errors += 1
                error = ET.SubElement(testcase, "error")
                error.set("message", "test failure")
                error.text = msg

        system_out = ET.SubElement(testcase, "system-out")
        system_out.text = string_cleanup(
            "\n".join(event.metadata.get("logs", "")), self.keep_restricted
        )

    def _check(self):
        if not os.path.exists(os.path.dirname(self.path)):
            raise OSError(
                2, "JUnitXML: Parent folder does not exist for file", self.path
            )
        if self.test_properties is not None:
            if not os.path.exists(self.test_properties_path):
                raise OSError(
                    2,
                    "JUnitXML: Properties file does not exist",
                    self.test_properties_path,
                )

    def stopTestRun(self, event):
        """Output xml tree to file"""
        self.tree.set("name", "nose2-junit")
        self.tree.set("errors", str(self.errors))
        self.tree.set("failures", str(self.failed))
        self.tree.set("skipped", str(self.skipped))
        self.tree.set("tests", str(self.numtests))
        self.tree.set("time", "%.3f" % event.timeTaken)

        self._check()
        self._include_test_properties()
        self._indent_tree(self.tree)

        output = ET.ElementTree(self.tree)
        output.write(self.path, encoding="utf-8")

    def _include_test_properties(self):
        """Include test properties in xml tree"""
        if self.test_properties is None:
            return

        props = {}
        with open(self.test_properties_path) as data:
            try:
                props = json.loads(data.read())
            except ValueError:
                raise ValueError(
                    "JUnitXML: could not decode file: '%s'" % self.test_properties_path
                )

        properties = ET.SubElement(self.tree, "properties")
        for key, val in props.items():
            prop = ET.SubElement(properties, "property")
            prop.set("name", key)
            prop.set("value", val)

    def _indent_tree(self, elem, level=0):
        """In-place pretty formatting of the ElementTree structure."""
        i = "\n" + level * "  "
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = i + "  "
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
            for e in elem:
                self._indent_tree(e, level + 1)
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
        else:
            if level and (not elem.tail or not elem.tail.strip()):
                elem.tail = i

    def _time(self):
        try:
            return time.time() - self._start
        except Exception:
            pass
        finally:
            self._start = None
        return 0

    def _iso_timestamp(self):
        # in some cases, `self._start` is still None when this runs, convert to 0
        # see: https://github.com/nose-devs/nose2/pull/505
        timestamp = self._start or 0
        return str(datetime.datetime.utcfromtimestamp(timestamp).isoformat())


#
# xml utility functions
#

# etree outputs XML 1.0 so the 1.1 Restricted characters are invalid.
# and there are no characters that can be given as entities aside
# form & < > ' " which ever have to be escaped (etree handles these fine)
ILLEGAL_RANGES = [
    (0x00, 0x08),
    (0x0B, 0x0C),
    (0x0E, 0x1F),
    (0xD800, 0xDFFF),
    (0xFFFE, 0xFFFF),
]
# 0xD800 thru 0xDFFF are technically invalid in UTF-8 but PY2 will encode
# bytes into these but PY3 will do a replacement

# Other non-characters which are not strictly forbidden but
# discouraged.
RESTRICTED_RANGES = [(0x7F, 0x84), (0x86, 0x9F), (0xFDD0, 0xFDDF)]
# check for a wide build
if sys.maxunicode > 0xFFFF:
    RESTRICTED_RANGES += [
        (0x1FFFE, 0x1FFFF),
        (0x2FFFE, 0x2FFFF),
        (0x3FFFE, 0x3FFFF),
        (0x4FFFE, 0x4FFFF),
        (0x5FFFE, 0x5FFFF),
        (0x6FFFE, 0x6FFFF),
        (0x7FFFE, 0x7FFFF),
        (0x8FFFE, 0x8FFFF),
        (0x9FFFE, 0x9FFFF),
        (0xAFFFE, 0xAFFFF),
        (0xBFFFE, 0xBFFFF),
        (0xCFFFE, 0xCFFFF),
        (0xDFFFE, 0xDFFFF),
        (0xEFFFE, 0xEFFFF),
        (0xFFFFE, 0xFFFFF),
        (0x10FFFE, 0x10FFFF),
    ]

ILLEGAL_REGEX_STR = (
    "[" + "".join([f"{chr(lo)}-{chr(hi)}" for (lo, hi) in ILLEGAL_RANGES]) + "]"
)
RESTRICTED_REGEX_STR = (
    "[" + "".join([f"{chr(lo)}-{chr(hi)}" for (lo, hi) in RESTRICTED_RANGES]) + "]"
)

_ILLEGAL_REGEX = re.compile(ILLEGAL_REGEX_STR, re.U)
_RESTRICTED_REGEX = re.compile(RESTRICTED_REGEX_STR, re.U)


def string_cleanup(string, keep_restricted=False):
    if not issubclass(type(string), str):
        string = str(string, encoding="utf-8", errors="replace")

    string = _ILLEGAL_REGEX.sub("\uFFFD", string)
    if not keep_restricted:
        string = _RESTRICTED_REGEX.sub("\uFFFD", string)

    return string



"""
This module provides decorators that assist the test author to write tests.
"""


def with_setup(setup):
    """
    A decorator that sets the :func:`setup` method to be executed before the test.

    It currently works only for function test cases.

    :param setup: The method to be executed before the test.
    :type setup: function
    """

    def decorator(testcase):
        testcase.setup = setup

        return testcase

    return decorator


def with_teardown(teardown):
    """
    A decorator that sets the :func:`teardown` method to be after before the test.

    It currently works only for function test cases.

    :param teardown: The method to be executed after the test.
    :type teardown: function
    """

    def decorator(testcase):
        testcase.tearDownFunc = teardown

        return testcase

    return decorator



from nose2.tests._common import FunctionalTestCase
from nose2.tools import such


class TestSuchDSL(FunctionalTestCase):
    def test_runs_example(self):
        proc = self.runIn("such", "-v", "--plugin=nose2.plugins.layers", "test_such")
        self.assertTestRunOutputMatches(proc, stderr="Ran 9 tests")
        self.assertEqual(proc.poll(), 0, proc.stderr.getvalue())

    def test_load_top_level_by_name(self):
        proc = self.runIn(
            "such",
            "-v",
            "--plugin=nose2.plugins.layers",
            "test_such.A system with complex setup.should do something",
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertEqual(proc.poll(), 0, proc.stderr.getvalue())

    def test_load_sublayer_test_by_name(self):
        proc = self.runIn(
            "such",
            "-v",
            "--plugin=nose2.plugins.layers",
            "test_such.having an expensive fixture." "should do more things",
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test")
        self.assertEqual(proc.poll(), 0, proc.stderr.getvalue())

    def test_regression_tests_with_the_same_having_description_under_different_fixtures_in_the_same_module_should_be_run(  # noqa
        self,
    ):
        proc = self.runIn(
            "such",
            "-v",
            "--plugin=nose2.plugins.layers",
            "test_regression_same_havings",
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 2 test")
        self.assertEqual(proc.poll(), 0, proc.stderr.getvalue())

    def test_teardown_fail(self):
        proc = self.runIn(
            "scenario/layers_with_errors",
            "--plugin=nose2.plugins.layers",
            "-v",
            "test_such_teardown_fail",
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 1 test in")
        self.assertTestRunOutputMatches(proc, stderr="ERROR: LayerSuite")
        self.assertTestRunOutputMatches(proc, stderr=r"FAILED \(errors=1\)")
        self.assertTestRunOutputMatches(proc, stderr="Bad Error in such tearDown")

    def test_setup_fail(self):
        proc = self.runIn(
            "scenario/layers_with_errors",
            "--plugin=nose2.plugins.layers",
            "-v",
            "test_such_setup_fail",
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 0 tests in")
        self.assertTestRunOutputMatches(proc, stderr="ERROR: LayerSuite")
        self.assertTestRunOutputMatches(proc, stderr=r"FAILED \(errors=1\)")
        self.assertTestRunOutputMatches(proc, stderr="Bad Error in such setUp!")

    def test_param_plugin_with_such(self):
        proc = self.runIn(
            "scenario/such_with_params",
            "--plugin=nose2.plugins.layers",
            "-v",
            "such_with_params",
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 6 tests in")
        self.assertTestRunOutputMatches(proc, stderr="OK")
        self.assertTestRunOutputMatches(proc, stderr="test 0000: should do bar:3")
        self.assertTestRunOutputMatches(
            proc, stderr="test 0001: should do bar and extra:3"
        )

    def test_such_without_layers_plugin(self):
        proc = self.runIn("such", "-v", "test_such_without_layers")
        self.assertTestRunOutputMatches(proc, stderr=r"FAILED")
        self.assertTestRunOutputMatches(
            proc, stderr=such.LAYERS_PLUGIN_NOT_LOADED_MESSAGE
        )

    def test_testsetup_on_higher_layer_with_test(self):
        proc = self.runIn(
            "scenario/layers_setups",
            "--plugin=nose2.plugins.layers",
            "-v",
            "higher_layer_testsetup_with_test",
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 3 tests in")
        self.assertTestRunOutputMatches(proc, stderr="OK")

    def test_testsetup_on_higher_layer(self):
        proc = self.runIn(
            "scenario/layers_setups",
            "--plugin=nose2.plugins.layers",
            "-v",
            "higher_layer_testsetup_no_test",
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 2 tests in")
        self.assertTestRunOutputMatches(proc, stderr="OK")

    def test_testsetup_on_higher_layer_3layers(self):
        proc = self.runIn(
            "scenario/layers_setups",
            "--plugin=nose2.plugins.layers",
            "-v",
            "higher_layer_testsetup_3layers",
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 2 tests in")
        self.assertTestRunOutputMatches(proc, stderr="OK")

    def test_setup_on_higher_layer(self):
        proc = self.runIn(
            "scenario/layers_setups",
            "--plugin=nose2.plugins.layers",
            "-v",
            "higher_layer_setup",
        )
        self.assertTestRunOutputMatches(proc, stderr="Ran 2 tests in")
        self.assertTestRunOutputMatches(proc, stderr="OK")

    def test_long_setup(self):
        proc = self.runIn(
            "such", "-v", "--plugin=nose2.plugins.layers", "test_such_timing"
        )
        self.assertTestRunOutputMatches(proc, stderr=r"Ran 2 tests in")
        self.assertTestRunOutputMatches(proc, stderr="OK")
